import json
import unittest

from common.settings import ENCODING, RESPONSE, ERROR, PRESENCE
from common.utils import send_message, receiving_message


class TestSocket:
    def __init__(self, test_message_dict):
        self.test_message_dict = test_message_dict
        self.test_message_bytes = None
        self.received_message = None

    def send(self, message_to_send):
        test_message_json = json.dumps(self.test_message_dict)
        self.test_message_bytes = test_message_json.encode(ENCODING)
        self.received_message = message_to_send

    def recv(self, max_len):
        test_message_json = json.dumps(self.test_message_dict)
        return test_message_json.encode(ENCODING)


class TestUtils(unittest.TestCase):
    test_message_send = {
        'action': PRESENCE,
        'time': 1,
        'user': {
            'account_name': 'guest__',
        }
    }
    test_message_recv_200 = {RESPONSE: 200}
    test_message_recv_400 = {
        RESPONSE: 400,
        ERROR: 'Bad Request'
    }

    def test_send_message(self):
        test_socket = TestSocket(self.test_message_send)
        send_message(test_socket, self.test_message_send)
        self.assertEqual(
            test_socket.test_message_bytes,
            test_socket.received_message)
        with self.assertRaises(Exception):
            send_message(test_socket, test_socket)

    def test_get_response(self):
        test_sock_ok = TestSocket(self.test_message_recv_200)
        test_sock_err = TestSocket(self.test_message_recv_400)
        self.assertEqual(
            receiving_message(test_sock_ok),
            self.test_message_recv_200)
        self.assertEqual(
            receiving_message(test_sock_err),
            self.test_message_recv_400)


if __name__ == '__main__':
    unittest.main()



import unittest

from client_app import generating_presence_message, pars_response_from_server
from common.settings import PRESENCE, RESPONSE, ERROR


class TestClient(unittest.TestCase):

    def test_presense(self):
        test_presense_dict = generating_presence_message()
        test_presense_dict['time'] = 1
        self.assertEqual(
            test_presense_dict, {
                'action': PRESENCE, 'time': 1, 'user': {
                    'account_name': 'guest'}})

    def test_pars_response_200(self):
        self.assertEqual(pars_response_from_server(
            {RESPONSE: 200}), '200 : OK')

    def test_pars_response_400(self):
        self.assertEqual(pars_response_from_server(
            {RESPONSE: 400, ERROR: 'Bad request'}), '400 : Bad request')

    def test_pars_response_error(self):
        self.assertRaises(
            ValueError, pars_response_from_server, {
                ERROR: 'Bad request'})


if __name__ == '__main__':
    unittest.main()



import time
import unittest

from common.settings import PRESENCE, RESPONSE, ERROR
from server_app import response_to_client


class TestServer(unittest.TestCase):
    response_error = {RESPONSE: 400, ERROR: 'Bad request'}

    def test_response_to_client_200(self):
        self.assertEqual(response_to_client({
            'action': PRESENCE,
            'time': time.time(),
            'user': {
                'account_name': 'guest',
            }
        }), {RESPONSE: 200})

    def test_response_to_client_not_guest(self):
        self.assertEqual(response_to_client({
            'action': PRESENCE,
            'time': time.time(),
            'user': {
                'account_name': 'guest__',
            }}), self.response_error)

    def test_response_to_client_no_action(self):
        self.assertEqual(response_to_client({
            'time': time.time(),
            'user': {
                'account_name': 'guest',
            }}), self.response_error)

    def test_response_to_client_no_time(self):
        self.assertEqual(response_to_client({
            'action': PRESENCE,
            'user': {
                'account_name': 'guest',
            }}), self.response_error)

    def test_response_to_client_no_user(self):
        self.assertEqual(response_to_client({
            'action': PRESENCE,
            'time': time.time(),
        }), self.response_error)

    def test_response_to_client_not_PRESENSE(self):
        self.assertEqual(response_to_client({
            'action': 123,
            'time': time.time(),
            'user': {
                'account_name': 'guest__',
            }}), self.response_error)


if __name__ == '__main__':
    unittest.main()



import logging

server_logger = logging.getLogger('server')


class Port:
    def __set__(self, instance, value):
        if not 1023 < value < 65536:
            server_logger.critical(
                f'Попытка запуска сервера с указанием неподходящего порта {value}. Допустимы адреса с 1024 до 65535.')
            exit(1)
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name


class Address:
    def __set__(self, instance, value):
        if value != '':
            value_list = value.split('.')

            if len(value_list) != 4:
                server_logger.critical(
                    f'Попытка запуска сервера с указанием неподходящего адреса {value}.')
                exit(1)
            for val in value_list:
                try:
                    val = int(val)
                except ValueError:
                    server_logger.critical(
                        f'Адрес сервера должен состоять из цифр.')
                    exit(1)
                if not 0 <= val <= 255:
                    server_logger.critical(
                        f'Попытка запуска сервера с указанием неподходящего адреса {value}. '
                        f'Значения октетов должны быть от 0 до 255.')
                    exit(1)

        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name



from common.settings import LOGGING_LEVEL
import logging.handlers
import os
import sys

# sys.path.append('../../')

# os.chdir(os.path.dirname(os.path.abspath(__file__)))
# PATH = os.path.dirname(os.path.abspath(__file__))
PATH = os.getcwd()
PATH = os.path.join(PATH, 'log/server.log')


server_log = logging.getLogger('server')
server_log.setLevel(LOGGING_LEVEL)

server_handler = logging.handlers.TimedRotatingFileHandler(
    PATH, encoding='utf_8', interval=1, when='D')

server_format = logging.Formatter(
    "%(asctime)-25s %(levelname)-9s %(filename)-20s %(message)s")
server_handler.setFormatter(server_format)

server_log.addHandler(server_handler)

if __name__ == '__main__':
    server_log.debug('отладочная информация')
    server_log.info('информационное сообщение')
    server_log.warning('предупреждение')
    server_log.error('ошибка')
    server_log.critical('критическая ошибка')



import datetime
import os
import threading

from sqlalchemy import create_engine, Table, Column, Integer, String, MetaData, ForeignKey, DateTime
from sqlalchemy.orm import mapper, sessionmaker, aliased
from sqlalchemy.sql import default_comparator


class ServerStorage:
    class Users:
        def __init__(self, username, hash_password):
            self.id = None
            self.username = username
            self.hash_password = hash_password
            self.public_key = None
            self.last_login = datetime.datetime.now()

    class ActiveUsers:
        def __init__(self, user_id, ip_address, port, login_time):
            self.id = None
            self.user_id = user_id
            self.ip_address = ip_address
            self.port = port
            self.login_time = login_time

    class LoginHistory:
        def __init__(self, username_id, date_time, ip_address, port):
            self.id = None
            self.username_id = username_id
            self.login_time = date_time
            self.ip_address = ip_address
            self.port = port

    class ContactList:
        def __init__(self, username_id, contact_id):
            self.id = None
            self.username_id = username_id
            self.contact_id = contact_id

    # история сообщений
    class MessageHistory:
        def __init__(self, username_id):
            self.id = None
            self.username_id = username_id
            self.number_messages_sent = 0
            self.number_messages_received = 0

    def __init__(self, path, database_lock):
        self.database_lock = database_lock
        # path = os.path.dirname(os.path.realpath(__file__))
        path = os.getcwd()
        filename = f'server_database.db3'
        self.database_engine = create_engine(
            f'sqlite:///{os.path.join(path, filename)}',
            echo=False,
            pool_recycle=7200,
            connect_args={
                'check_same_thread': False})

        self.metadata = MetaData()
        users_table = Table('Users', self.metadata,
                            Column('id', Integer, primary_key=True),
                            Column('username', String, unique=True),
                            Column('hash_password', String),
                            Column('public_key', String),
                            Column('last_login', DateTime)
                            )

        active_users_table = Table(
            'Active_users', self.metadata, Column(
                'id', Integer, primary_key=True), Column(
                'user_id', ForeignKey('Users.id'), unique=True), Column(
                'ip_address', String), Column(
                    'port', Integer), Column(
                        'login_time', DateTime))

        login_history_table = Table(
            'Login_history', self.metadata, Column(
                'id', Integer, primary_key=True), Column(
                'username_id', ForeignKey('Users.id')), Column(
                'login_time', DateTime), Column(
                    'ip_address', String), Column(
                        'port', String))

        contact_list_table = Table(
            'Contact_list', self.metadata, Column(
                'id', Integer, primary_key=True), Column(
                'username_id', ForeignKey('Users.id')), Column(
                'contact_id', ForeignKey('Users.id')))

        message_history_table = Table(
            'Message_history', self.metadata, Column(
                'id', Integer, primary_key=True), Column(
                'username_id', ForeignKey('Users.id')), Column(
                'number_messages_sent', Integer), Column(
                    'number_messages_received', Integer))

        self.metadata.create_all(self.database_engine)

        mapper(self.Users, users_table)
        mapper(self.ActiveUsers, active_users_table)
        mapper(self.LoginHistory, login_history_table)
        mapper(self.ContactList, contact_list_table)
        mapper(self.MessageHistory, message_history_table)

        Session = sessionmaker(bind=self.database_engine)
        self.session = Session()

        # Когда устанавливаем соединение, очищаем таблицу активных
        # пользователей
        self.session.query(self.ActiveUsers).delete()
        self.session.commit()

    def user_login(self, username, ip_address, port, key):
        print(username, ip_address, port)

        rez = self.session.query(self.Users).filter_by(username=username)

        # обновляем время последнего входа
        if rez.count():
            user = rez.first()
            user.last_login = datetime.datetime.now()
            if user.public_key != key:
                user.public_key = key
        # Если нет, то создаздаём нового пользователя
        else:
            user = self.Users(username)
            self.session.add(user)
            self.session.commit()
            user_in_history = self.MessageHistory(user.id)
            self.session.add(user_in_history)
            self.session.commit()

        # Создаем запись в таблицу активных пользователей.
        new_active_user = self.ActiveUsers(
            user.id, ip_address, port, datetime.datetime.now())
        self.session.add(new_active_user)
        self.session.commit()

        history = self.LoginHistory(
            user.id, datetime.datetime.now(), ip_address, port)
        self.session.add(history)
        self.session.commit()

    # Функция фиксирующая отключение пользователя
    def user_logout(self, username):
        user = self.session.query(
            self.Users).filter_by(
            username=username).first()

        # Удаляем запись из таблицы ActiveUsers
        self.session.query(
            self.ActiveUsers).filter_by(
            user_id=user.id).delete()
        self.session.commit()

    def users_list(self):
        query = self.session.query(
            self.Users.username,
            self.Users.last_login,
        )
        return query.all()

    # Функция возвращает список активных пользователей
    def active_users_list(self):
        with self.database_lock:
            query = self.session.query(
                self.Users.username,
                self.ActiveUsers.ip_address,
                self.ActiveUsers.port,
                self.ActiveUsers.login_time
            ).join(self.Users)

        return query.all()

    # Функция возвращающая историю входов по пользователю или всем
    # пользователям
    def login_history(self, username=None):
        query = self.session.query(self.Users.username,
                                   self.LoginHistory.login_time,
                                   self.LoginHistory.ip_address,
                                   self.LoginHistory.port
                                   ).join(self.Users)

        if username:
            query = query.filter(self.Users.username == username)
        return query.all()

    # Функция возвращает список контактов пользователя
    def contact_list(self, username=None):
        alias1 = aliased(self.Users)
        alias2 = aliased(self.Users)
        query = self.session.query(alias1.username,
                                   alias2.username
                                   ) \
            .select_from(self.ContactList) \
            .join(alias1, alias1.id == self.ContactList.username_id) \
            .join(alias2, alias2.id == self.ContactList.contact_id)

        if username:
            query = query.filter(alias1.username == username)
        return [contact[1] for contact in query.all()]

    # Функция добавляет запись в БД о новом контакте contact_name, для
    # user_name
    def add_contact(self, user_name, contact_name):
        # Получаем id user'а и contact'а
        user_id = self.session.query(
            self.Users).filter_by(
            username=user_name).first().id
        contact_id = self.session.query(
            self.Users).filter_by(
            username=contact_name).first().id
        # Если такая запись уже есть в таблице завершаем работу функции
        if self.session.query(
                self.ContactList).filter_by(
                username_id=user_id).count():
            return
        # если нет то добавляем
        contact = self.ContactList(user_id, contact_id)
        self.session.add(contact)
        self.session.commit()

    # Функция удаляет запись из БД о контакте contact_name, для user_name
    def delete_contact(self, user_name, contact_name):
        # Получаем id user'а и contact'а
        user_id = self.session.query(
            self.Users).filter_by(
            username=user_name).first().id
        contact_id = self.session.query(
            self.Users).filter_by(
            username=contact_name).first().id
        # Проверяем наличие удаляемой записи в таблице
        if self.session.query(
                self.ContactList).filter_by(
                username_id=user_id,
                contact_id=contact_id).count():
            self.session.query(
                self.ContactList).filter_by(
                username_id=user_id,
                contact_id=contact_id).delete()
            self.session.commit()
        else:
            return

    def message_history_add(self, sender, recipient):
        sender_id = self.session.query(
            self.Users).filter_by(
            username=sender).first().id
        recipient_id = self.session.query(
            self.Users).filter_by(
            username=recipient).first().id

        sender_row = self.session.query(
            self.MessageHistory).filter_by(
            username_id=sender_id).first()

        recipient_row = self.session.query(
            self.MessageHistory).filter_by(
            username_id=recipient_id).first()
        if recipient_row:
            recipient_row.number_messages_received += 1
        if sender_row:
            sender_row.number_messages_sent += 1

        self.session.commit()

    def message_history(self):
        query = self.session.query(
            self.Users.username,
            self.Users.last_login,
            self.MessageHistory.number_messages_sent,
            self.MessageHistory.number_messages_received
        ).join(self.Users)

        return query.all()

    def check_user(self, username):
        if self.session.query(self.Users).filter_by(username=username).count():
            return True
        else:
            return False

    def user_registration(self, username, hash_password):
        user = self.Users(username, hash_password)
        self.session.add(user)
        self.session.commit()
        history = self.MessageHistory(user.id)
        self.session.add(history)
        self.session.commit()

    def get_hash(self, username):
        query = self.session.query(
            self.Users).filter_by(
            username=username).first()
        return query.hash_password

    def get_pubkey(self, username):
        user = self.session.query(
            self.Users).filter_by(
            username=username).first()
        return user.public_key


if __name__ == '__main__':
    test_db = ServerStorage()
    # выполняем 'подключение' пользователя
    test_db.user_login('client_1', '192.168.1.4', 8888)
    test_db.user_login('client_2', '192.168.1.5', 7777)
    # выводим список кортежей - активных пользователей
    print('Список активных пользователей:')
    print(test_db.active_users_list())
    # выполянем 'отключение' пользователя
    test_db.user_logout('client_1')
    # выводим список активных пользователей
    print(test_db.active_users_list())
    # запрашиваем историю входов по пользователю
    print('История входов пользователя client_1:')
    print(test_db.login_history('client_1'))
    # выводим список известных пользователей
    print('Cписок известных пользователей')
    print(test_db.users_list())
    print('Список контактов:')
    print(test_db.contact_list())
    # print(test_db.contact_list('client_2'))
    print('Список контакттов после уделения:')
    print(test_db.delete_contact('client_1', 'client_2'))
    test_db.add_contact('client_1', 'client_2')
    print(test_db.message_history())



import sys
import threading

from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QMainWindow, QAction, qApp, QApplication, QLabel, QTableView, QDialog, QPushButton, \
    QLineEdit, QFileDialog

database_lock = threading.Lock()


class MainWindow(QMainWindow):
    def __init__(self, database):
        super().__init__()
        self.database = database
        self.initUI()

    def initUI(self):
        exitAction = QAction('Выход', self)
        exitAction.setShortcut('Ctrl+Q')
        exitAction.triggered.connect(qApp.quit)

        self.refresh_button = QAction('Обновить список', self)

        self.config_btn = QAction('Настройки сервера', self)

        self.show_history_button = QAction('История клиентов', self)

        self.statusBar()

        self.toolbar = self.addToolBar('MainBar')
        self.toolbar.addAction(exitAction)
        self.toolbar.addAction(self.refresh_button)
        self.toolbar.addAction(self.show_history_button)
        self.toolbar.addAction(self.config_btn)

        self.setFixedSize(800, 600)
        self.setWindowTitle('Server_app')

        self.label = QLabel('Список подключённых клиентов:', self)
        self.label.setFixedSize(240, 15)
        self.label.move(10, 25)

        self.active_clients_table = QTableView(self)
        self.active_clients_table.move(10, 45)
        self.active_clients_table.setFixedSize(780, 400)

        self.timer = QTimer()
        self.timer.timeout.connect(self.gui_create_model)
        self.timer.start(1000)

        self.refresh_button.triggered.connect(self.gui_create_model)
        self.show_history_button.triggered.connect(self.create_statistics_window)
        self.config_btn.triggered.connect(self.create_config_window)

        self.show()

    def gui_create_model(self):
        list_users = self.database.active_users_list()
        list = QStandardItemModel()
        list.setHorizontalHeaderLabels(
            ['Имя Клиента', 'IP Адрес', 'Порт', 'Время подключения'])
        for row in list_users:
            user, ip, port, time = row
            user = QStandardItem(user)
            user.setEditable(False)
            ip = QStandardItem(ip)
            ip.setEditable(False)
            port = QStandardItem(str(port))
            port.setEditable(False)

            time = QStandardItem(str(time.replace(microsecond=0)))
            time.setEditable(False)
            list.appendRow([user, ip, port, time])
        self.active_clients_table.setModel(list)
        self.active_clients_table.resizeColumnsToContents()
        self.active_clients_table.resizeRowsToContents()

    def create_config_window(self):
        global config_window
        config_window = ConfigWindow()

    def create_statistics_window(self):
        global statistics_window
        statistics_window = HistoryWindow(self.database)


class HistoryWindow(QDialog):
    def __init__(self, database):
        super().__init__()
        self.database = database
        self.initUI()

    def initUI(self):
        self.setWindowTitle('Статистика клиентов')
        self.setFixedSize(600, 700)
        self.setAttribute(Qt.WA_DeleteOnClose)

        self.close_button = QPushButton('Закрыть', self)
        self.close_button.move(250, 650)
        self.close_button.clicked.connect(self.close)

        self.history_table = QTableView(self)
        self.history_table.move(10, 10)
        self.history_table.setFixedSize(580, 620)

        self.create_stat_model()
        self.show()

    def create_stat_model(self):
        hist_list = self.database.message_history()

        list = QStandardItemModel()
        list.setHorizontalHeaderLabels(
            ['Имя Клиента', 'Последний раз входил', 'Сообщений отправлено', 'Сообщений получено'])
        for row in hist_list:
            user, last_seen, sent, recvd = row
            user = QStandardItem(user)
            user.setEditable(False)
            last_seen = QStandardItem(str(last_seen.replace(microsecond=0)))
            last_seen.setEditable(False)
            sent = QStandardItem(str(sent))
            sent.setEditable(False)
            recvd = QStandardItem(str(recvd))
            recvd.setEditable(False)
            list.appendRow([user, last_seen, sent, recvd])
        self.history_table.setModel(list)
        self.history_table.resizeColumnsToContents()
        self.history_table.resizeRowsToContents()


class ConfigWindow(QDialog):
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        self.setFixedSize(365, 260)
        self.setWindowTitle('Настройки сервера')

        self.db_path_label = QLabel('Путь до файла базы данных: ', self)
        self.db_path_label.move(10, 10)
        self.db_path_label.setFixedSize(240, 15)

        self.db_path_edit = QLineEdit(self)
        self.db_path_edit.setFixedSize(250, 20)
        self.db_path_edit.move(10, 30)
        self.db_path_edit.setReadOnly(True)

        self.db_path_button = QPushButton('Обзор...', self)
        self.db_path_button.move(275, 28)

        def open_file_dialog():
            global dialog
            dialog = QFileDialog(self)
            path = dialog.getExistingDirectory()
            path = path.replace('/', '\\')
            self.db_path_edit.insert(path)

        self.db_path_button.clicked.connect(open_file_dialog)

        self.db_file_label = QLabel('Имя файла базы данных: ', self)
        self.db_file_label.move(10, 68)
        self.db_file_label.setFixedSize(180, 15)

        self.db_file_edit = QLineEdit(self)
        self.db_file_edit.move(200, 66)
        self.db_file_edit.setFixedSize(150, 20)

        self.port_label = QLabel('Номер порта для соединений:', self)
        self.port_label.move(10, 108)
        self.port_label.setFixedSize(180, 15)

        self.port_edit = QLineEdit(self)
        self.port_edit.move(200, 108)
        self.port_edit.setFixedSize(150, 20)

        self.ip_label = QLabel('С какого IP принимаем соединения:', self)
        self.ip_label.move(10, 148)
        self.ip_label.setFixedSize(180, 15)

        self.ip_label_note = QLabel(
            ' оставьте это поле пустым, чтобы\n принимать соединения с любых адресов.',
            self)
        self.ip_label_note.move(10, 168)
        self.ip_label_note.setFixedSize(500, 30)

        self.ip_edit = QLineEdit(self)
        self.ip_edit.move(200, 148)
        self.ip_edit.setFixedSize(150, 20)

        self.save_button = QPushButton('Сохранить', self)
        self.save_button.move(190, 220)

        self.close_button = QPushButton('Закрыть', self)
        self.close_button.move(275, 220)
        self.close_button.clicked.connect(self.close)

        self.show()


if __name__ == '__main__':
    app = QApplication(sys.argv)
    ex = MainWindow()
    ex.statusBar().showMessage('Test Statusbar Message')
    test_list = QStandardItemModel(ex)
    test_list.setHorizontalHeaderLabels(
        ['Имя Клиента', 'IP Адрес', 'Порт', 'Время подключения'])
    test_list.appendRow(
        [QStandardItem('Client_1'), QStandardItem('ip_1'), QStandardItem('port_1')])
    test_list.appendRow(
        [QStandardItem('Client_2'), QStandardItem('ip_2'), QStandardItem('port_2')])
    ex.active_clients_table.setModel(test_list)
    ex.active_clients_table.resizeColumnsToContents()
    app.exec_()
    print('END')

    # app = QApplication(sys.argv)
    # message = QMessageBox
    # dial = ConfigWindow()

    app.exec_()



import inspect
import logging
import sys

if 'client' in sys.argv[0]:
    logger = logging.getLogger('client')
else:
    logger = logging.getLogger('server')


def log(func):
    def decorated(*args, **kwargs):
        logger.info(
            f'Функция {func.__name__} вызвана из функции {inspect.stack()[1][3]}, с аргументами {args} {kwargs}',
            stacklevel=2)
        # print(f'Функция {func.__name__} вызвана из функции {inspect.stack()[1][3]}, с аргументами {args} {kwargs}')
        return func(*args, **kwargs)

    return decorated


class Log():
    def __call__(self, func):
        def decorated(*args, **kwargs):
            logger.info(
                f'Функция {func.__name__} вызвана из функции {inspect.stack()[1][3]}, с аргументами {args} {kwargs}',
                stacklevel=2)
            # print(f'Функция {func.__name__} вызвана из функции {inspect.stack()[1][3]}, с аргументами {args} {kwargs}')
            return func(*args, **kwargs)

        return decorated


if __name__ == '__main__':
    @Log()
    def temp(a, b):
        print(a, b)

    def main():
        temp(1, 3)

    main()



import logging

ENCODING = 'utf-8'

USER = 'user'
PRESENCE = 'presence'
RESPONSE = 'response'
ERROR = 'error'
ACTION = 'action'
TIME = 'time'
ACCOUNT_NAME = 'account_name'
MESSAGE_TEXT = 'message_text'
MESSAGE = 'message'
SENDER = 'sender'
EXIT = 'exit'
DELETE_CONTACT = 'remove'
ADD_CONTACT = 'add'
USERS_REQUEST = 'get_users'
CONTACT_LIST = 'get_contacts'
DATA = 'data'
PUBLIC_KEY = 'pubkey'
REGISTRATION = 'reg'
PASSWORD = 'passwd'
DATA_LIST = 'data_list'
DESTINATION = 'to'
PUBLIC_KEY_REQUEST = 'pubkey_request'

DEFAULT_PORT = 7777
DEFAULT_IP_ADDRESS = '127.0.0.1'

MAX_PACKAGE_LENGTH = 1024

LOGGING_LEVEL = logging.DEBUG

SERVER_DATABASE = 'sqlite:///server_database.db3'

RESPONSE_200 = {RESPONSE: 200}

RESPONSE_202 = {
    RESPONSE: 202,
    'data_list': None
}

RESPONSE_400 = {
    RESPONSE: 400,
    ERROR: None
}

RESPONSE_511 = {
    RESPONSE: 511,
    DATA: None
}



import json

from common.settings import ENCODING, MAX_PACKAGE_LENGTH


# Кодирует сообщение в байты и отправляет его
def send_message(socket, message):
    json_message = json.dumps(message)
    message_bytes = json_message.encode(ENCODING)
    socket.send(message_bytes)


# Принимает сообщение в байтах и декодирует его
def receiving_message(socket):
    response_bytes = socket.recv(MAX_PACKAGE_LENGTH)
    if isinstance(response_bytes, bytes):
        response_json = response_bytes.decode(ENCODING)
        response = json.loads(response_json)
        if isinstance(response, dict):
            return response
        raise ValueError
    raise ValueError



import dis


class ServerVerifier(type):
    def __init__(self, clsname, bases, clsdict):

        methods = []
        attrs = []

        for func in clsdict:
            try:
                instructions = dis.get_instructions(clsdict[func])
            except TypeError:
                pass
            else:
                for instruction in instructions:
                    # print(instruction)
                    if instruction.opname == 'LOAD_METHOD':
                        if instruction.argval not in methods:
                            methods.append(instruction.argval)
                    elif instruction.opname == 'LOAD_ATTR':
                        if instruction.argval not in attrs:
                            attrs.append(instruction.argval)
        # print(methods)
        # print(attrs)

        if 'connect' in methods:
            raise TypeError(
                'Использование метода connect недопустимо в серверном классе')

        if not ('SOCK_STREAM' in attrs and 'AF_INET' in attrs):
            raise TypeError('Некорректная инициализация сокета.')

        super().__init__(clsname, bases, clsdict)


class ClientVerifier(type):
    def __init__(self, clsname, bases, clsdict):
        methods = []
        for func in clsdict:
            try:
                instructions = dis.get_instructions(clsdict[func])
            except TypeError:
                pass
            else:
                for instruction in instructions:
                    # print(instruction)
                    if instruction.opname == 'LOAD_GLOBAL':
                        if instruction.argval not in methods:
                            methods.append(instruction.argval)
        # print(methods)
        for command in ('accept', 'listen', 'socket'):
            if command in methods:
                raise TypeError(
                    'В классе обнаружено использование запрещённого метода')
        if 'receiving_message' in methods or 'send_message' in methods:
            pass
        else:
            raise TypeError(
                'Отсутствуют вызовы функций, работающих с сокетами.')
        super().__init__(clsname, bases, clsdict)



import argparse
import binascii
import configparser
import hmac
import json
import logging
import os
import select
import sys
import threading
import socket

from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QApplication, QMessageBox

import log.configs.server_log_config
from common.descriptors import Port, Address
from common.metaclasses import ServerVerifier
from common.settings import PRESENCE, RESPONSE, ERROR, ACTION, ACCOUNT_NAME, TIME, MESSAGE, MESSAGE_TEXT, \
    SENDER, USER, RESPONSE_200, RESPONSE_202, RESPONSE_400, ADD_CONTACT, DELETE_CONTACT, CONTACT_LIST, USERS_REQUEST, \
    RESPONSE_511, DATA, PUBLIC_KEY, PASSWORD, DESTINATION, PUBLIC_KEY_REQUEST
from common.utils import receiving_message, send_message
from server.server_database import ServerStorage
from server.server_gui import MainWindow, HistoryWindow, ConfigWindow

server_logger = logging.getLogger('server')
database_lock = threading.Lock()


class Server(threading.Thread, metaclass=ServerVerifier):

    #"""Основной класс сервера. Работает в отдельном потоке."""

    address = Address()
    port = Port()

    def __init__(self, address, port, database):
        self.clients = []
        self.messages = []
        # словарь имя: сокет.
        self.client_names = dict()

        self.listen_sockets = None
        self.error_sockets = None

        self.address = address
        self.port = port
        self.database = database
        super().__init__()

    def run(self):
        """
        Метод основного цикла работы сервера.
        Инициализирует сокет сервера.
        Устанавливает соедение с клиентом.
        Принимает сообщения от клиентов.
        """

        # self.address, self.port = self.arg_parser()

        self.socket_server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket_server.bind((self.address, self.port))
        self.socket_server.settimeout(0.5)
        # server_logger.info(f'Установлено соедение с клиентом {self.address}')
        self.socket_server.listen()

        while True:
            try:
                client, client_address = self.socket_server.accept()
            except OSError:
                pass
            else:
                server_logger.info(
                    f'Установлено соедение с клиентом {client_address}')
                self.clients.append(client)

            recv_list = []
            # send_list = []
            # err_list = []

            try:
                if self.clients:
                    recv_list, self.listen_sockets, self.error_sockets = select.select(
                        self.clients, self.clients, [], 0)
            except OSError as err:
                server_logger.error(f'Ошибка работы с сокетами: {err}')
            except ValueError as err:
                server_logger.error(f'Ошибка работы с сокетами: {err}')

            if recv_list:
                for client_with_message in recv_list:
                    # print(client_with_message)
                    try:
                        self.pars_message_from_client(
                            receiving_message(client_with_message), client_with_message)

                    except (OSError, json.JSONDecodeError, TypeError) as err:
                        server_logger.info(
                            f'Клиент {client_with_message.getpeername()} отключился от сервера.',
                            exc_info=err)
                        self.remove_client(client_with_message)

    # @Log()
    @staticmethod
    def arg_parser(default_port, default_address):
        """Парсер аргументов коммандной строки."""

        parser = argparse.ArgumentParser()
        parser.add_argument('-p', default=default_port, type=int, nargs='?')
        parser.add_argument('-a', default=default_address, nargs='?')
        namespace = parser.parse_args(sys.argv[1:])
        listen_addr = namespace.a
        listen_port = namespace.p

        return listen_addr, listen_port

    # @Log()
    # def pars_message_from_client(self, message, messages_list, client,
    # clients, names):
    def pars_message_from_client(self, message, client):
        """
        Метод разбора сообщения от клиента.

        :param message (dict): Сообщение от клиента.
        :param client (str): Имя клиента.
        """

        # global new_connection
        server_logger.debug(f'Разбор сообщения от клиента : {message}')

        if ACTION in message and message[ACTION] == PRESENCE and TIME in message \
                and USER in message:
            users_list = [user[0] for user in self.database.users_list()]
            # print(users_list)
            # если клиента нет в БД (в списке известных пользователей) -
            # регистрируем
            if message[USER][ACCOUNT_NAME] not in users_list:
                with database_lock:
                    self.database.user_registration(
                        message[USER][ACCOUNT_NAME], message[USER][PASSWORD])

            self.authorization(message, client)

        elif ACTION in message and message[ACTION] == MESSAGE and 'to' in message and TIME in message \
                and SENDER in message and MESSAGE_TEXT in message and self.client_names[message[SENDER]] == client:
            if message['to'] in self.client_names:
                with database_lock:
                    self.database.message_history_add(
                        message[SENDER], message['to'])
                self.send_message_to_client(message)
                try:
                    send_message(client, RESPONSE_200)
                except OSError:
                    self.remove_client(client)
            else:
                response = RESPONSE_400
                response[ERROR] = 'Пользователь не зарегистрирован на сервере.'
                try:
                    send_message(client, response)
                except OSError:
                    pass
            return

        elif ACTION in message and message[ACTION] == 'exit' and ACCOUNT_NAME in message:
            with database_lock:
                self.database.user_logout(message[ACCOUNT_NAME])
            server_logger.info(
                f'Клиент {message[ACCOUNT_NAME]} корректно отключился от сервера.')
            self.clients.remove(self.client_names[message[ACCOUNT_NAME]])
            self.client_names[message[ACCOUNT_NAME]].close()
            del self.client_names[message[ACCOUNT_NAME]]
            # with conflag_lock:
            #     new_connection = True
            # return

        elif ACTION in message and message[ACTION] == MESSAGE and \
                'to' in message and TIME in message \
                and SENDER in message and MESSAGE_TEXT in message:
            self.messages.append(message)
            with database_lock:
                self.database.message_history_add(message[SENDER], message['to'])
            try:
                send_message(client, RESPONSE_200)
            except OSError:
                self.remove_client(client)
            return

        elif ACTION in message and message[ACTION] == CONTACT_LIST and USER in message:
            # self.client_names[message[USER]] == client:
            response = RESPONSE_202
            with database_lock:
                response['data_list'] = self.database.contact_list(message[USER])
            try:
                send_message(client, response)
            except OSError:
                self.remove_client(client)

        elif ACTION in message and message[ACTION] == ADD_CONTACT and ACCOUNT_NAME in message and USER in message:
            # and self.client_names[message[USER]] == client:
            with database_lock:
                self.database.add_contact(message[USER], message[ACCOUNT_NAME])
            try:
                send_message(client, RESPONSE_200)
            except OSError:
                self.remove_client(client)

        elif ACTION in message and message[ACTION] == DELETE_CONTACT and ACCOUNT_NAME in message and USER in message:
            # and self.client_names[message[USER]] == client:
            with database_lock:
                self.database.delete_contact(message[USER], message[ACCOUNT_NAME])
            try:
                send_message(client, RESPONSE_200)
            except OSError:
                self.remove_client(client)

        elif ACTION in message and message[ACTION] == USERS_REQUEST and ACCOUNT_NAME in message:
            # and self.client_names[message[ACCOUNT_NAME]] == client:
            response = RESPONSE_202
            with database_lock:
                response['data_list'] = [user[0]
                                     for user in self.database.users_list()]
            try:
                send_message(client, response)
            except OSError:
                self.remove_client(client)

        elif ACTION in message and message[ACTION] == PUBLIC_KEY_REQUEST and ACCOUNT_NAME in message:
            response = RESPONSE_511
            with database_lock:
                response[DATA] = self.database.get_pubkey(message[ACCOUNT_NAME])
            if response[DATA]:
                try:
                    send_message(client, response)
                except OSError:
                    self.remove_client(client)
            else:
                response = RESPONSE_400
                response[ERROR] = 'Нет публичного ключа для данного пользователя'
                try:
                    send_message(client, response)
                except OSError:
                    self.remove_client(client)
        else:
            response = RESPONSE_400
            response[ERROR] = 'Запрос некорректен.'
            try:
                send_message(client, response)
            except OSError:
                self.remove_client(client)
            return

    def authorization(self, message, client):
        """
        Метод реализующий авторизацию пользователей.

        :param message (dict): Сообщение с запросом на авторизацию.
        :param client (socket): Сокет клиента.
        """
        with database_lock:
            check_user = self.database.check_user(message[USER][ACCOUNT_NAME])
        if not check_user:
            response = RESPONSE_400
            response[ERROR] = 'Пользователь не зарегистрирован.'
            try:
                server_logger.debug(
                    f'Пользователь не зарегистрирован. {response}')
                send_message(client, response)
            except OSError:
                pass
            self.clients.remove(client)
            client.close()
        elif message[USER][ACCOUNT_NAME] in self.client_names.keys():
            response = RESPONSE_400
            response[ERROR] = 'Имя пользователя уже занято.'
            try:
                send_message(client, response)
            except OSError:
                server_logger.debug('OS Error')
            self.clients.remove(client)
            client.close()
        else:
            message_auth = RESPONSE_511
            # message_auth[DATA] = os.urandom(64).decode('base-64')
            random_str = binascii.hexlify(os.urandom(64))
            message_auth[DATA] = random_str.decode('ascii')
            # send_message(client, message_auth)

            with database_lock:
                key = self.database.get_hash(
                    message[USER][ACCOUNT_NAME]).encode('ascii')
            hash = hmac.new(key, random_str, 'MD5')
            digest = hash.digest()

            try:
                send_message(client, message_auth)
                response = receiving_message(client)
            except OSError as err:
                server_logger.debug('Ошибка авторизации, data:', exc_info=err)
                client.close()
                return
            client_digest = binascii.a2b_base64(response[DATA])

            # print(response)
            if RESPONSE in response and response[RESPONSE] == 511 and hmac.compare_digest(
                    digest, client_digest):
                self.client_names[message[USER][ACCOUNT_NAME]] = client
                client_ip, client_port = client.getpeername()
                try:
                    send_message(client, RESPONSE_200)
                except OSError:
                    self.remove_client(message[USER][ACCOUNT_NAME])

                with database_lock:
                    self.database.user_login(message[USER][ACCOUNT_NAME],
                                         client_ip,
                                         client_port,
                                         message[USER][PUBLIC_KEY])
            else:
                response = RESPONSE_400
                response[ERROR] = 'Неверный пароль.'
                try:
                    send_message(client, response)
                except OSError:
                    pass
                self.clients.remove(client)
                client.close()
            # with conflag_lock:
            #     new_connection = True

    # @Log()
    def send_message_to_client(self, message):
        """
        Метод пересылающий сообщение от одного клиента другому.

        :param message (dict): Сообщение. Содержит информацию об отправителе и получателе, а также само сообщение.
        """
        if message[DESTINATION] in self.client_names and self.client_names[message[DESTINATION]
        ] in self.listen_sockets:
            try:
                send_message(self.client_names[message[DESTINATION]], message)
                server_logger.info(
                    f'Отправлено сообщение пользователю {message[DESTINATION]} от пользователя {message[SENDER]}.')
            except OSError:
                self.remove_client(message[DESTINATION])
        elif message[DESTINATION] in self.client_names and self.client_names[
            message[DESTINATION]] not in self.listen_sockets:
            server_logger.error(
                f'Связь с клиентом {message[DESTINATION]} была потеряна. Соединение закрыто, доставка невозможна.')
            self.remove_client(self.client_names[message[DESTINATION]])
        else:
            server_logger.error(
                f'Пользователь {message[DESTINATION]} не зарегистрирован на сервере, отправка сообщения невозможна.')

    #     if message['to'] in self.client_names and self.client_names[message['to']] in self.listen_sockets:
    #         send_message(self.client_names[message['to']], message)
    #         server_logger.info(f'Отправлено сообщение пользователю {message["to"]} '
    #                            f'от пользователя {message[SENDER]}.')
    #         self.database.add_contact(message[SENDER], message["to"])
    #     elif message['to'] in self.client_names and self.client_names[message['to']] not in self.listen_sockets:
    #         raise ConnectionError
    #     else:
    #         server_logger.error(
    #             f'Пользователь {message["to"]} не зарегистрирован на сервере, '
    #             f'отправка сообщения невозможна.')

    def remove_client(self, client_name):
        """
        Метод отключающий клиента от сервера.

        :param client_name (socket): сокет клиента"""
        server_logger.info(
            f'Клиент {client_name.getpeername()} отключился от сервера.')
        for name in self.client_names:
            if self.client_names[name] == client_name:
                with database_lock:
                    self.database.user_logout(name)
                del self.client_names[name]
                break
        self.clients.remove(client_name)
        client_name.close()


def print_help():
    print('Поддерживаемые комманды:')
    print('users - список известных пользователей')
    print('contact - список контактов пользователей')
    print('connected - список подключенных пользователей')
    print('loghist - история входов пользователя')
    print('exit - завершение работы сервера.')
    print('help - вывод справки по поддерживаемым командам')


def main():
    config = configparser.ConfigParser()

    # dir_path = os.path.dirname(os.path.realpath(__file__))
    dir_path = os.getcwd()
    config_path = os.path.join(dir_path, 'server/server.ini')
    # config.read(f"{dir_path}/{'server.ini'}")
    config.read(config_path)

    # Загрузка параметров командной строки, если нет параметров, то задаём
    # значения по умоланию.
    address, port = Server.arg_parser(
        config['SETTINGS']['Default_port'], config['SETTINGS']['Listen_Address'])

    # Инициализация базы данных
    database = ServerStorage(os.path.join(
        config['SETTINGS']['Database_path'],
        config['SETTINGS']['Database_file']),
        database_lock)

    server = Server(address, port, database)
    server.daemon = True
    server.start()

    server_app = QApplication(sys.argv)
    main_window = MainWindow(database)

    # main_window.statusBar().showMessage('Server working')
    # main_window.active_clients_table.setModel(gui_create_model(database))
    # main_window.active_clients_table.resizeColumnsToContents()
    # main_window.active_clients_table.resizeRowsToContents()

    # def list_update():
    #     # global new_connection
    #     # if new_connection:
    #         # main_window.active_clients_table.setModel(
    #         #     gui_create_model(database))
    #     main_window.active_clients_table.resizeColumnsToContents()
    #     main_window.active_clients_table.resizeRowsToContents()
    #         # with conflag_lock:
    #         #     new_connection = False
    #
    # def window_statistics():
    #     global stat_window
    #     stat_window = HistoryWindow(database)
    #     # stat_window.history_table.setModel(create_stat_model(database))
    #     stat_window.history_table.resizeColumnsToContents()
    #     stat_window.history_table.resizeRowsToContents()
    #     stat_window.show()
    #
    # def window_config():
    #     global config_window
    #     # Создаём окно и заносим в него текущие параметры
    #     config_window = ConfigWindow()
    #     config_window.db_path_edit.insert(config['SETTINGS']['Database_path'])
    #     config_window.db_file_edit.insert(config['SETTINGS']['Database_file'])
    #     config_window.port_edit.insert(config['SETTINGS']['Default_port'])
    #     config_window.ip_edit.insert(config['SETTINGS']['Listen_Address'])
    #     config_window.save_button.clicked.connect(save_server_config)
    #
    # def save_server_config():
    #     global config_window
    #     message = QMessageBox()
    #     config['SETTINGS']['Database_path'] = config_window.db_path_edit.text()
    #     config['SETTINGS']['Database_file'] = config_window.db_file_edit.text()
    #     try:
    #         port = int(config_window.port_edit.text())
    #     except ValueError:
    #         message.warning(config_window, 'Ошибка', 'Порт должен быть числом')
    #     else:
    #         config['SETTINGS']['Listen_Address'] = config_window.ip_edit.text()
    #         if 1023 < port < 65536:
    #             config['SETTINGS']['Default_port'] = str(port)
    #             print(port)
    #             with open('server.ini', 'w') as conf:
    #                 config.write(conf)
    #                 message.information(
    #                     config_window, 'OK', 'Настройки успешно сохранены!')
    #         else:
    #             message.warning(
    #                 config_window,
    #                 'Ошибка',
    #                 'Порт должен быть от 1024 до 65536')

    # # Таймер, обновляющий список клиентов 1 раз в секунду
    # timer = QTimer()
    # timer.timeout.connect(list_update)
    # timer.start(1000)

    # # Связываем кнопки с процедурами
    # main_window.refresh_button.triggered.connect(list_update)
    # main_window.show_history_button.triggered.connect(window_statistics)
    # main_window.config_btn.triggered.connect(window_config)

    server_app.exec_()


if __name__ == '__main__':
    main()
